import doctest
import difflib
import json

def compare_dicts(d1, d2):
    d = difflib.Differ()
    return list(d.compare(
        json.dumps(d1, indent=4, sort_keys=True).splitlines(),
        json.dumps(d2, indent=4, sort_keys=True).splitlines()
    ))

def doctests(module, tests):
    """
    A helper function to combine unittest discovery with doctest suites.

    Args:
        module: The module to add doctests from.
        tests: The existing TestSuite discovered by unittest.

    Returns:
        A combined TestSuite with doctests added.
    """
    tests.addTests(doctest.DocTestSuite(module))
    return tests


import unittest
import utils
import json
import pycocowriter.utils

class TestAttrDict(unittest.TestCase):

    example_dict = {
        'meta': {
            'skiprows': 0
        },
        'filename': 3,
        'label': 10, 
        'keypoints': [
            {
                'name': 'head',
                'x': 13,
                'y': 14
            },
            {
                'name': 'tail',
                'x': 15,
                'y': 16
            }
        ],
        'keypoint_skeleton': [1,2]
    }

    def test_attrdict(self):
        attr_dict = pycocowriter.utils.AttrDict(TestAttrDict.example_dict)
        self.assertEqual(attr_dict.meta.skiprows, 0)
        self.assertEqual(attr_dict.keypoints[1].name, 'tail')
        self.assertEqual(attr_dict.keypoint_skeleton[1], 2)

def load_tests(loader, tests, ignore):
    return utils.doctests(pycocowriter.utils, tests)

if __name__ == '__main__':
    unittest.main()


import json
from .coco import *

def coco_merge(*coco_dicts: dict, info: COCOInfo = COCOInfo()) -> dict:
    """
    Merge COCO annotations.

    adapted from pyodi https://github.com/Gradiant/pyodi under MPL 2.0

    Parameters
    ----------
    *coco_files: dicts
        dicts from coco files as e.g. read by json.load
    
    Returns
    -------
    coco_data: dict
        coco data for merged output as dict
    """

    categories = []
    images = []
    annotations = []
    licenses = []

    license_map = {}
    category_map = {}
    image_map = {}

    for data in coco_dicts:
        
        cat_id_map = {}
        for new_cat in data["categories"]:
            label = new_cat["name"]
            if label in category_map:
                cat_id_map[new_cat["id"]] = category_map[label]
            else:
                new_id = len(categories) + 1
                cat_id_map[new_cat["id"]] = new_id
                new_cat["id"] = new_id
                categories.append(new_cat)
                category_map[label] = new_id

        license_id_map = {}
        for new_license in data["licenses"]:
            license_name = new_license["name"]
            if license_name in license_map:
                license_id_map[new_license["id"]] = license_map[license_name]
            else:
                new_id = len(licenses) + 1
                license_id_map[new_license["id"]] = new_id
                new_license["id"] = new_id
                licenses.append(new_license)
                license_map[license_name] = new_id

        image_id_map = {}
        for new_image in data["images"]:
            file_name = new_image["file_name"]
            if file_name in image_map:
                image_id_map[new_image["id"]] = image_map[file_name]
            else:
                new_id = len(images) + 1
                image_id_map[new_image["id"]] = new_id
                new_image["id"] = new_id
                if "license_id" in new_image:
                    new_image["license_id"] = license_id_map[new_image["license_id"]]
                images.append(new_image)
                image_map[file_name] = new_id

        for new_annotation in data["annotations"]:
            new_id = len(annotations) + 1
            new_annotation["id"] = new_id
            new_annotation["category_id"] = cat_id_map[new_annotation["category_id"]]
            new_annotation["image_id"] = image_id_map[new_annotation["image_id"]]
            annotations.append(new_annotation)

    return {
        "annotations": annotations,
        "images": images,
        "info": info.to_dict(),
        "categories": categories,
        "licenses": licenses
    }


from collections.abc import Iterable
import json
import numpy as np

class AttrDict(dict):
    '''
    This class allows javascript-like attribute access for dictionaries.
    Pass in a nested dictionary and recursively access members as attributes.

    Examples
    --------
    >>> attr_dict = AttrDict({'foo': [{'bar': [1,2]}]})
    >>> attr_dict.foo[0].bar[1] == 2
    True
    '''
    def __init__(self, some_dict:dict):
        super().__init__(some_dict)
        self.__dict__ = {
            k: _to_attrdict(v) for k,v in some_dict.items()
        }

def _to_attrdict(el) -> AttrDict:
    '''
    This method facilitates recursive application of AttrDict.
    Probably you don't want to call this method on its own -
    instead, instantiate an `AttrDict`

    Parameters
    ----------
    el: any
        element to recursively convert to AttrDict.
    
    Returns
    -------
    any
        the element, possibly converted to an AttrDict depending on type 
    '''
    # cast dictionaries at AttrDict
    if isinstance(el, dict):
        return AttrDict(el)
    # return strings
    if isinstance(el, str):
        return el
    # try to recurse into other iterable types
    if isinstance(el, Iterable):
        return type(el)((
            _to_attrdict(v) for v in el
        ))
    # return simple types unchanged
    return el

class NPEncoder(json.JSONEncoder):
    '''
    json encoder class used to convert numpy types into plain python types during json.dumps.
    
    Examples
    --------
    >>> json.dumps({'array': np.array([1,2,3])}, cls=NPEncoder)
    '{"array": [1, 2, 3]}'

    '''
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NPEncoder, self).default(obj)
    
def skiprows(iterable: Iterable, n: int) -> Iterable:
    '''
    skips the first n rows of iterable.  returns iterable as a convenience

    Parameters
    ----------
    iterable: Iterable
        the iterable to skip rows of
    n: int
        the number of rows to skip

    Returns
    -------
    iterable: Iterable
        the original iterable, but now with n rows having been skipped
    '''
    for i in range(n):
        next(iterable)
    return iterable


from ultralytics.data.converter import convert_coco
from pycocotools.coco import COCO
import os
import yaml
from pathlib import Path
import shutil

''' 
Constants that are expected/used by yolo or the ultralytics
conversion tools
''' 
IMAGE_DIR = 'images'
LABEL_DIR = 'labels'
ULTRALYTICS_COCO_CONVERSION_DIR = 'coco_converted'

def coco2yoloyaml(coco_file_dir: str, destination: str) -> None:
    '''
    Construct a YOLO-format yaml metadata file from a 
    collection of COCO-format annotation json files.

    # Train/val/test sets as 1) dir: path/to/imgs, 2) file: path/to/imgs.txt, or 3) list: [path/to/imgs1, path/to/imgs2, ..]
    path: ../datasets/coco128 # dataset root dir
    train: images/train2017 # train images (relative to 'path') 128 images
    val: images/train2017 # val images (relative to 'path') 128 images
    test: # test images (optional)

    # Classes (80 COCO classes)
    names:
        0: person
        1: bicycle
        2: car
        # ...
        77: teddy bear
        78: hair drier
        79: toothbrush

    Parameters
    ----------
    coco_file_dir: str
        The location of the COCO json files.  If "test" is in the name
        of the file, it is assumed that this is for a test dataset.
        If "val" is in the name of the file, it is assumed that this is for
        a validation dataset.  Anything else is assumed to be training data 
    destination: str
        Where to write the resulting yaml file
    '''
    result = {
        'path': destination,
        'names': {}
    }
    os.makedirs(destination, exist_ok=True)
    for coco_file in sorted(Path(coco_file_dir).resolve().glob("*.json")):
        coco = COCO(coco_file)
        basename = os.path.splitext(os.path.basename(coco_file))[0]
        img_dir = os.path.join(basename, IMAGE_DIR)
        for k in coco.cats:
            # The classes in the coco files need to be the same!
            result['names'][k-1] = coco.cats[k]['name']
        if 'val' in basename:
            result['val'] = img_dir
        elif 'test' in basename:
            result['test'] = img_dir
        else:
            result['train'] = img_dir
    if 'val' not in result:
        result['val'] = result['train']
    with open(os.path.join(destination, 'train.yaml'),'w') as f:
        yaml.dump(result, f)
        
def download_coco_images(coco_file_dir: str, destination: str) -> None:
    '''
    download the images referenced in the "coco_url" of the coco files
    in a directory.

    Parameters
    ----------
    coco_file_dir: str
        The location of the COCO json files.
    destination: str
        The root directory to which the images should be writ.
    '''
    os.makedirs(destination, exist_ok=True)
    for coco_file in sorted(Path(coco_file_dir).resolve().glob("*.json")):
        coco = COCO(coco_file)
        basename = os.path.splitext(os.path.basename(coco_file))[0]
        img_dir = os.path.join(destination, basename, IMAGE_DIR)
        coco.download(img_dir)

def rename_label_paths(coco_file_dir: str, destination: str) -> None:
    '''
    depending on the `destination` to which we download the images,
    our labels need to be updated to point to this location.

    Parameters
    ----------
    coco_file_dir: str
        The location of the COCO json files.
    destination: str
        The root directory to which the images should be writ.
    '''
    os.makedirs(destination, exist_ok=True)
    for coco_file in sorted(Path(coco_file_dir).resolve().glob("*.json")):
        basename = os.path.splitext(os.path.basename(coco_file))[0]
        # where the ultralytics script stores the labels
        label_path = os.path.join(
            ULTRALYTICS_COCO_CONVERSION_DIR, LABEL_DIR, basename)
        # where they had ought to be given our image destination
        dest_label_path = os.path.join(
            destination, basename, LABEL_DIR)
        try:
            shutil.rmtree(dest_label_path)
        except FileNotFoundError:
            pass    # Fail silently if the directory doesn't exist
        shutil.move(label_path, dest_label_path)
        shutil.rmtree(ULTRALYTICS_COCO_CONVERSION_DIR)

def coco2yolo(coco_file_dir: str, destination: str, 
              use_segments: bool = False, use_keypoints: bool = False):
    '''
    Convert a COCO-format annotation dataset and metadata 
    into YOLO-format annotations and metadata.  Downloads the
    `coco_url` files into an appropriate location.

    Defaults to bounding-box style annotations.  If the dataset
    uses segmentation or keypoints, these details are not automatically
    inferred, and this needs to be toggled.  At most one of
    `use_segments` or `use_keypoints` should be True

    Parameters
    ----------
    coco_file_dir: str
        The location of the COCO json files.
    destination: str
        The root directory to which the images should be writ.
    use_segments: bool
        Whether to attempt to read segmentation data from the COCO
        files, and create a YOLO segmentation dataset
    use_keypoints: bool        
        Whether to attempt to read keypoint data from the COCO
        files, and create a YOLO keypoint dataset
    '''
    convert_coco(
        coco_file_dir, 
        use_segments=use_segments, 
        use_keypoints=use_keypoints, 
        cls91to80=False
    )
    rename_label_paths(coco_file_dir, destination)
    coco2yoloyaml(coco_file_dir, destination)
    download_coco_images(coco_file_dir, destination)

#if __name__ == '__main__':
#    coco2yolo('../data/raw/oceaneyes_example', '../data/raw/oceaneyes_example/yolo_example_data', use_segments=False, use_keypoints=False)



import json
import jsonschema
from . import coco, utils
import csv
from collections.abc import Sequence, Iterable


def parse_csv(config: dict, filename: str) -> tuple[
    list[coco.COCOImage], list[coco.COCOAnnotation], list[coco.COCOCategory]]:
    """Helper method to open a csv file and pass it row-by-row into the COCO builder

    Parameters
    ----------
    config : dict
        a dictionary conforming to the Iterable2COCOConfig.SCHEMA
    filename : str
        a csv file to be read and converted to COCO

    Returns
    -------
    images : list[COCOImage]
        a list of COCOImage reflecting the images from the csv file
    annotations : list[COCOAnnotation]
        a list of COCOAnnotation reflecting the annotations from the csv file
    categories : list[COCOCategory]
        a list of COCOCategory reflecting the categories of the annotations
    """
    csv2coco = Iterable2COCO(Iterable2COCOConfig(config))
    with open(filename) as f:
        reader = csv.reader(f)
        images, annotations, categories = csv2coco.parse(reader)
    return images, annotations, categories


def bbox_tlbr2xywh(bbox: tuple[int, int, int, int]) -> tuple[int, int, int, int]:
    '''
    Convert a bounding box in "top left, bottom right" format 
    to a bounding box in "top left, width height" format

    Parameters
    ----------
    bbox: tuple[int,int,int,int]
        a four-tuple of [top_left_x, top_left_y, bottom_right_x, bottom_right_y]

    Returns
    -------
    bbox: tuple[int,int,int,int]    
        a four-tuple of [top_left_x, top_left_y, width, height]
    '''
    return (bbox[0], bbox[1], bbox[2] - bbox[0], bbox[3] - bbox[1])


class Iterable2COCOConfig(utils.AttrDict):
    '''
    This class validates a configuration to convert a "flat" iterable type into COCO.
    Because COCO has complex nested and optional types, it is not possible to have a 
    "one-size-fits-all" flat iterable to COCO conversion.  This configuration tells
    the converter which fields are present, and in which columns they are located.
    This class exists only to validate a dict as a valid configuration.

    Parameters
    ----------
    config: dict
        a configuration dictionary adhering to SCHEMA.  Set up this way to be read from .json

    Attributes
    ----------
    SCHEMA: dict
        a jsonschema representing a valid configuration
    '''

    # TODO: This is the 'anything' schema.  Update to reflect the actual rules
    SCHEMA = {}

    def __init__(self, config: dict):
        self._validate_config(config)
        super().__init__(config)

    def _validate_config(self, config: dict):
        jsonschema.validate(config, Iterable2COCOConfig.SCHEMA)


class IterableBBoxParser(object):
    '''
    This class is to help parse bounding boxes from "row" data.  Sometimes these data
    are in different formats, so this class is intended to assist in dealing with these nuances

    Parameters
    ----------
    config: Iterable2COCOConfig
        this configuration should define how (and if) bounding boxes are present in each row
    '''

    def __init__(self, config: Iterable2COCOConfig):
        self.config = config
        self._init_bbox_method()

    def _init_bbox_tlbr(self):
        '''
        configures the `get_bbox` method to get bounding boxes in "top left, width/height" 
        format, given bounding boxes in "top left, bottom right" format
        '''
        self.get_bbox = self._get_bbox_tlbr
        self.bbox_cols = [
            self.config.bbox_tlbr.tlx,
            self.config.bbox_tlbr.tly,
            self.config.bbox_tlbr.brx,
            self.config.bbox_tlbr.bry,
        ]

    def _init_bbox_xywh(self):
        '''
        configures the `get_bbox` method to get bounding boxes in "top left, width/height" 
        format, given bounding boxes in "top left, width/height" format
        '''
        self.get_bbox = self._get_bbox_xywh
        self.bbox_cols = [
            self.config.bbox_xywh.x,
            self.config.bbox_xywh.y,
            self.config.bbox_xywh.w,
            self.config.bbox_xywh.h,
        ]

    def _init_bbox_method(self):
        '''
        dispatches configuration of the `get_bbox` method depending on the 
        contents of the configuration file.
        '''
        if 'bbox_tlbr' in self.config:
            self._init_bbox_tlbr()
        elif 'bbox_xywh' in self.config:
            self._init_bbox_xywh()

    def get_bbox(self, row: Sequence) -> list[int, int, int, int] | None:
        '''
        this method gets overwritten in __init__ if the config has a bbox option

        Parameters
        ----------
        row: Sequence
            a row, e.g. from a csv.  The bounding box should be in some columns of this row
            as defined in the configuration

        Returns
        -------
        bbox: list[int,int,int,int]
            the bounding box as [top_left_x, top_left_y, width, height]
        '''
        return None

    def _get_bbox_tlbr(self, row: Sequence) -> list[int, int, int, int]:
        '''
        gets a bounding box in "top left, width/height" format given bounding box subsetted from a  
        row from, e.g. a csv.  The subset of columns in the input row should be defined in the config, 
        and should contain a bounding box in "top left, bottom right" format.

        Parameters
        ----------
        row: Sequence
            a row, e.g. from a csv.  The bounding box should be in some columns of this row
            as defined in the configuration

        Returns
        -------
        bbox: list[int,int,int,int]
            the bounding box as [top_left_x, top_left_y, width, height]
        '''
        return bbox_tlbr2xywh([int(float(row[i])) for i in self.bbox_cols])

    def _get_bbox_xywh(self, row: Sequence) -> list[int, int, int, int]:
        '''
        gets a bounding box in "top left, width/height" format given bounding box subsetted from a  
        row from, e.g. a csv.  The subset of columns in the input row should be defined in the config, 
        and should contain a bounding box in "top left, width/height" format.

        Parameters
        ----------
        row: Sequence
            a row, e.g. from a csv.  The bounding box should be in some columns of this row
            as defined in the configuration

        Returns
        -------
        bbox: list[int,int,int,int]
            the bounding box as [top_left_x, top_left_y, width, height]
        '''
        return [int(float(row[i])) for i in self.bbox_cols]


class IterableKeypointParser(object):
    '''
    This class is to help parse keypoints from "row" data.  Sometimes these data
    are in different formats, so this class is intended to assist in dealing with these nuances

    Parameters
    ----------
    config: Iterable2COCOConfig
        this configuration should define how (and if) bounding boxes are present in each row
    '''
    FULLY_VISIBLE_COCO_KEYPOINT = 2

    def __init__(self, config: Iterable2COCOConfig):
        self.config = config

    def keypoint_config(self) -> tuple[list[str], list[list[int]]]:
        '''
        get the keypoint layout from the configuration file.  the coco keypoint layout is
        ['kpname1', 'kpname2', ...]
        and also a skeleton
        [edge1, edge2, ...]
        where edges are two-tuples of 1-indexed indexes of keypoints.  For example, if keypoints are:
        ['hip', 'knee', 'ankle'],
        the skeleton would be:
        [[1,2],[2,3]] because the hip has an edge with the knee, and the knee has an edge to the ankle.

        both of these items should be defined in the configuration

        TODO: we only support ONE keypoint structure per configuration right now....
        if you have multiple possible keypoint structures
        e.g. hands and also human poses, then we need to rework this to be more general

        Returns
        -------
        keypoints: list[str]
            the list of keypoint names
        skeleton: list[list[int]]
            the skeleton corresponding to the keypoint names
        '''
        if 'keypoints' not in self.config:
            return None, None
        return (
            [
                keypoint.name for keypoint in self.config.keypoints
            ],
            self.config.keypoint_skeleton
        )

    def get_keypoints(self, row: Sequence) -> list[int]:
        '''
        get keypoints from a "flat" row using expected indices in the row of keypoints 
        defined in self.config

        Parameters
        ----------
        row: Sequence
            a row, e.g. from a csv.  The keypoints should be in some columns of this row
            as defined in the configuration

        Returns
        -------
        keypoints: list[int]
            keypoint locations in form [x1,y1,v1,x2,y2,v2,....] where x,y are the location and v is the
            "visibility" according to the COCO docs
        '''
        if 'keypoints' not in self.config:
            return None
        return sum(
            [
                [
                    int(float(row[keypoint.x])),
                    int(float(row[int(keypoint.y)])),
                    IterableKeypointParser.FULLY_VISIBLE_COCO_KEYPOINT if 'visibility' not in keypoint else int(float(row[keypoint.visibility]))
                ]
                for keypoint in self.config.keypoints
            ],
            []
        )


class Iterable2COCO(object):
    '''
    class providing methods for parsing a "flat" iterable (e.g. a csv) of annotations into COCO
    format.  Each "row" should contain things such as the image filename, 
    and the annotation information (e.g a bounding box or keypoints and a category label)

    Parameters
    ----------
    config: Iterable2COCOConfig
        a configuration dictionary detailing which columns in each row correspond to various COCO
        features such as filename and bounding box coordinates.

    Attributes
    ----------
    bbox_parser: IterableBBoxParser
        a helper for parsing bounding boxes and the concomitant configuration
    keypoint_parser: IterableKeypointParser
        a helper for parsing keypoints and the concomitant configuration
    '''

    def __init__(self, config: Iterable2COCOConfig):
        self.config = config
        self.bbox_parser = IterableBBoxParser(config)
        self.keypoint_parser = IterableKeypointParser(config)

    def _get_scalar(self, field: str, row: Sequence):
        '''
        get a single value from a row given the field name in the configuration

        Parameters
        ----------
        field: str
            the field name as expected in the configuration, and should point to a column index.
            viz. self.config[field] should provide an index into row
        row: Sequence
            an indexable "row" e.g. from a csv file

        Returns
        -------
        scalar: any
           a single value expected for that field. 
        '''
        if field not in self.config:
            return None
        return row[self.config[field]]

    def parse(self, row_iterable: Iterable[Sequence]) -> tuple[
        list[coco.COCOImage], list[coco.COCOAnnotation], list[coco.COCOCategory]
    ]:
        '''
        parse an iterable of rows (e.g. from a csv file) containing image annotation information into
        COCO format.

        Parameters
        ----------
        row_iterable: Iterable[Sequence]
            an iterable of rows containing annotation information

        Returns
        -------
        images: list[COCOImage]
            a list of all unique images listed in the iterable, in COCO format
        annotations: list[COCOAnnotation]
            a list of all annotations listed in the iterable, correctly indexed
            against the images and categories lists
        categories: list[COCOCategory]
            a list of all unique categories listed in the iterable, in COCO format        
        '''
        categories = coco.COCOCategories()
        images = coco.COCOImages()
        annotations = []
        if 'meta' in self.config and 'skiprows' in self.config.meta:
            utils.skiprows(row_iterable, self.config.meta.skiprows)
        keypoint_names, keypoint_skeleton = self.keypoint_parser.keypoint_config()
        for row in row_iterable:
            bbox = self.bbox_parser.get_bbox(row)
            keypoints = self.keypoint_parser.get_keypoints(row)
            filename = self._get_scalar('filename', row)
            width = self._get_scalar('width', row)
            height = self._get_scalar('height', row)
            label = self._get_scalar('label', row)
            images.add(filename, width, height)
            categories.add(label, keypoint_names, keypoint_skeleton)
            annotations.append(
                coco.COCOAnnotation(
                    images.image_map[filename],
                    len(annotations),
                    categories.category_map[label],
                    bbox=bbox,
                    keypoints=keypoints
                )
            )
        return images.images, annotations, categories.categories



import unittest
from pycocowriter.coco import COCOData, COCOInfo
import utils
import json
import pycocowriter.csv2coco
import difflib

class TestCSV2COCO(unittest.TestCase):

    example_config = {
        'meta': {
            'skiprows': 2
        },
        'filename': 3,
        'label': 10,
        'keypoints': [
            {
                'name': 'head',
                'x': 13,
                'y': 14
            },
            {
                'name': 'tail',
                'x': 15,
                'y': 16
            }
        ],
        'bbox_tlbr': {
            'tlx': 4,
            'tly': 5,
            'brx': 6,
            'bry': 7
        },
        'keypoint_skeleton': [1, 2]
    }

    example_csv = [
        ['err'] * 17,
        ['err'] * 17,
        ['err', 'err', 'err', 'tests/static/example1.png', 10, 11, 20, 21, 'err', 'err', 'label1', 'err', 'err', 1, 2, 3, 4, 'err'],
        ['err', 'err', 'err', 'tests/static/example2.png', 100, 110, 200, 210, 'err', 'err', 'label1', 'err', 'err', 10, 20, 70, 80, 'err'],
        ['err', 'err', 'err', 'tests/static/example1.png', 1000, 1100, 2000, 2100, 'err', 'err', 'label2', 'err', 'err', 100, 200, 300, 400, 'err'],
    ]

    example_csv_as_coco = {
        'info': {},
        'images': [
            {
                'id': 1, 
                'file_name': 'tests/static/example1.png', 
                'width': 1, 
                'height': 1
            },
            {
                'id': 2, 
                'file_name': 'tests/static/example2.png', 
                'width': 1, 
                'height': 1
            }
        ],
        'annotations': [
            {
                'image_id': 1,
                'id': 0,
                'category_id': 1,
                'bbox': (10, 11, 10, 10),
                'area': 100,
                'iscrowd': 0,
                'keypoints': [1, 2, 2, 3, 4, 2],
                'num_keypoints': 2
            },
            {
                'image_id': 2,
                'id': 1,
                'category_id': 1,
                'bbox': (100, 110, 100, 100),
                'area': 10000,
                'iscrowd': 0,
                'keypoints': [10, 20, 2, 70, 80, 2],
                'num_keypoints': 2},
            {
                'image_id': 1,
                'id': 2,
                'category_id': 2,
                'bbox': (1000, 1100, 1000, 1000),
                'area': 1000000,
                'iscrowd': 0,
                'keypoints': [100, 200, 2, 300, 400, 2],
                'num_keypoints': 2
            }
        ],
        'licenses': [],
        'categories': [
            {
                'name': 'label1',
                'id': 1,
                'keypoints': ['head', 'tail'],
                'skeleton': [1, 2]
            },
            {
                'name': 'label2',
                'id': 2,
                'keypoints': ['head', 'tail'],
                'skeleton': [1, 2]
            }
        ]
    }

    def test_csv_parser(self):
        csv2coco_parser = pycocowriter.csv2coco.Iterable2COCO(
            pycocowriter.csv2coco.Iterable2COCOConfig(TestCSV2COCO.example_config)
        )
        images, annotations, categories = csv2coco_parser.parse(iter(TestCSV2COCO.example_csv))
        coco_data = COCOData(
            COCOInfo(),
            images,
            annotations,
            [],
            categories
        )
        coco_dict = coco_data.to_dict()
        '''
        print(json.dumps(coco_dict, indent=4))
        print(json.dumps(TestCSV2COCO.example_csv_as_coco, indent=4))
        print(json.dumps(
            utils.compare_dicts(
                coco_dict,
                TestCSV2COCO.example_csv_as_coco
            ), indent=4))
        '''
        self.assertDictEqual(coco_dict, TestCSV2COCO.example_csv_as_coco)


class TestCSV2COCONoKeypoints(unittest.TestCase):

    example_config = {
        'meta': {
            'skiprows': 2
        },
        'filename': 3,
        'label': 10,
        'bbox_tlbr': {
            'tlx': 4,
            'tly': 5,
            'brx': 6,
            'bry': 7
        }
    }

    example_csv = [
        ['err'] * 17,
        ['err'] * 17,
        ['err', 'err', 'err', 'tests/static/example1.png', 10, 11, 20, 21, 'err', 'err', 'label1', 'err', 'err', 1, 2, 3, 4, 'err'],
        ['err', 'err', 'err', 'tests/static/example2.png', 100, 110, 200, 210, 'err', 'err', 'label1', 'err', 'err', 10, 20, 70, 80, 'err'],
        ['err', 'err', 'err', 'tests/static/example1.png', 1000, 1100, 2000, 2100, 'err', 'err', 'label2', 'err', 'err', 100, 200, 300, 400, 'err'],
    ]

    example_csv_as_coco = {
        'info': {},
        'images': [
            {
                'id': 1, 
                'file_name': 'tests/static/example1.png', 
                'width': 1, 
                'height': 1
            },
            {
                'id': 2, 
                'file_name': 'tests/static/example2.png', 
                'width': 1, 
                'height': 1
            }
        ],
        'annotations': [
            {
                'image_id': 1,
                'id': 0,
                'category_id': 1,
                'bbox': (10, 11, 10, 10),
                'area': 100,
                'iscrowd': 0
            },
            {
                'image_id': 2,
                'id': 1,
                'category_id': 1,
                'bbox': (100, 110, 100, 100),
                'area': 10000,
                'iscrowd': 0
            },
            {
                'image_id': 1,
                'id': 2,
                'category_id': 2,
                'bbox': (1000, 1100, 1000, 1000),
                'area': 1000000,
                'iscrowd': 0
            }
        ],
        'licenses': [],
        'categories': [
            {
                'name': 'label1',
                'id': 1
            },
            {
                'name': 'label2',
                'id': 2
            }
        ]
    }

    def test_csv_parser(self):
        csv2coco_parser = pycocowriter.csv2coco.Iterable2COCO(
            pycocowriter.csv2coco.Iterable2COCOConfig(TestCSV2COCONoKeypoints.example_config)
        )
        images, annotations, categories = csv2coco_parser.parse(iter(TestCSV2COCONoKeypoints.example_csv))
        coco_data = COCOData(
            COCOInfo(),
            images,
            annotations,
            [],
            categories
        )
        coco_dict = coco_data.to_dict()
        self.assertDictEqual(coco_dict, TestCSV2COCONoKeypoints.example_csv_as_coco)


def load_tests(loader, tests, ignore):
    return utils.doctests(pycocowriter.csv2coco, tests)


if __name__ == '__main__':
    unittest.main()



import json
import unittest

from pycocowriter.cocomerge import coco_merge
import utils

class TestCocoMerge(unittest.TestCase):

    example_coco_1 = {
        'info': {},
        'images': [
            {
                'id': 1, 
                'file_name': 'tests/static/example1.png', 
                'width': 1, 
                'height': 1
            },
            {
                'id': 2, 
                'file_name': 'tests/static/example2.png', 
                'width': 1, 
                'height': 1,
                'license_id': 1
            }
        ],
        'annotations': [
            {
                'image_id': 1,
                'id': 0,
                'category_id': 1,
                'bbox': (10, 11, 10, 10),
                'area': 100,
                'iscrowd': 0,
                'keypoints': [1, 2, 2, 3, 4, 2],
                'num_keypoints': 2
            },
            {
                'image_id': 2,
                'id': 1,
                'category_id': 1,
                'bbox': (100, 110, 100, 100),
                'area': 10000,
                'iscrowd': 0,
                'keypoints': [10, 20, 2, 70, 80, 2],
                'num_keypoints': 2},
            {
                'image_id': 1,
                'id': 2,
                'category_id': 2,
                'bbox': (1000, 1100, 1000, 1000),
                'area': 1000000,
                'iscrowd': 0,
                'keypoints': [100, 200, 2, 300, 400, 2],
                'num_keypoints': 2
            }
        ],
        'licenses': [
            {
                'name': 'license1',
                'url': 'example.com',
                'id': 1
            }
        ],
        'categories': [
            {
                'name': 'label1',
                'id': 1,
                'keypoints': ['head', 'tail'],
                'skeleton': [1, 2]
            },
            {
                'name': 'label2',
                'id': 2,
                'keypoints': ['head', 'tail'],
                'skeleton': [1, 2]
            }
        ]
    }

    example_coco_2 = {
        'info': {},
        'images': [
            {
                'id': 1, 
                'file_name': 'tests/static/example3.png', 
                'width': 1, 
                'height': 1
            },
            {
                'id': 2, 
                'file_name': 'tests/static/example1.png', 
                'width': 1, 
                'height': 1,
                'license_id': 1
            }
        ],
        'annotations': [
            {
                'image_id': 1,
                'id': 0,
                'category_id': 1,
                'bbox': (11, 12, 11, 11),
                'area': 100,
                'iscrowd': 0,
                'keypoints': [1, 2, 2, 3, 4, 2],
                'num_keypoints': 2
            },
            {
                'image_id': 2,
                'id': 1,
                'category_id': 1,
                'bbox': (101, 111, 101, 101),
                'area': 10000,
                'iscrowd': 0,
                'keypoints': [10, 20, 2, 70, 80, 2],
                'num_keypoints': 2},
            {
                'image_id': 1,
                'id': 2,
                'category_id': 2,
                'bbox': (1001, 1101, 1001, 1001),
                'area': 1000000,
                'iscrowd': 0,
                'keypoints': [100, 200, 2, 300, 400, 2],
                'num_keypoints': 2
            }
        ],
        'licenses': [
            {
                'name': 'license2',
                'url': 'example.com',
                'id': 1
            }
        ],
        'categories': [
            {
                'name': 'label4',
                'id': 1,
                'keypoints': ['head', 'tail'],
                'skeleton': [1, 2]
            },
            {
                'name': 'label1',
                'id': 2,
                'keypoints': ['head', 'tail'],
                'skeleton': [1, 2]
            }
        ]
    }

    examples_merged = {
        "annotations": [
            {
                "image_id": 1,
                "id": 1,
                "category_id": 1,
                "bbox": (
                    10,
                    11,
                    10,
                    10
                ),
                "area": 100,
                "iscrowd": 0,
                "keypoints": [
                    1,
                    2,
                    2,
                    3,
                    4,
                    2
                ],
                "num_keypoints": 2
            },
            {
                "image_id": 2,
                "id": 2,
                "category_id": 1,
                "bbox": (
                    100,
                    110,
                    100,
                    100
                ),
                "area": 10000,
                "iscrowd": 0,
                "keypoints": [
                    10,
                    20,
                    2,
                    70,
                    80,
                    2
                ],
                "num_keypoints": 2
            },
            {
                "image_id": 1,
                "id": 3,
                "category_id": 2,
                "bbox": (
                    1000,
                    1100,
                    1000,
                    1000
                ),
                "area": 1000000,
                "iscrowd": 0,
                "keypoints": [
                    100,
                    200,
                    2,
                    300,
                    400,
                    2
                ],
                "num_keypoints": 2
            },
            {
                "image_id": 3,
                "id": 4,
                "category_id": 3,
                "bbox": (
                    11,
                    12,
                    11,
                    11
                ),
                "area": 100,
                "iscrowd": 0,
                "keypoints": [
                    1,
                    2,
                    2,
                    3,
                    4,
                    2
                ],
                "num_keypoints": 2
            },
            {
                "image_id": 1,
                "id": 5,
                "category_id": 3,
                "bbox": (
                    101,
                    111,
                    101,
                    101
                ),
                "area": 10000,
                "iscrowd": 0,
                "keypoints": [
                    10,
                    20,
                    2,
                    70,
                    80,
                    2
                ],
                "num_keypoints": 2
            },
            {
                "image_id": 3,
                "id": 6,
                "category_id": 1,
                "bbox": (
                    1001,
                    1101,
                    1001,
                    1001
                ),
                "area": 1000000,
                "iscrowd": 0,
                "keypoints": [
                    100,
                    200,
                    2,
                    300,
                    400,
                    2
                ],
                "num_keypoints": 2
            }
        ],
        "images": [
            {
                "id": 1,
                "file_name": "tests/static/example1.png",
                "width": 1,
                "height": 1
            },
            {
                "id": 2,
                "file_name": "tests/static/example2.png",
                "width": 1,
                "height": 1,
                "license_id": 1
            },
            {
                "id": 3,
                "file_name": "tests/static/example3.png",
                "width": 1,
                "height": 1
            }
        ],
        "info": {},
        "categories": [
            {
                "name": "label1",
                "id": 1,
                "keypoints": [
                    "head",
                    "tail"
                ],
                "skeleton": [
                    1,
                    2
                ]
            },
            {
                "name": "label2",
                "id": 2,
                "keypoints": [
                    "head",
                    "tail"
                ],
                "skeleton": [
                    1,
                    2
                ]
            },
            {
                "name": "label4",
                "id": 3,
                "keypoints": [
                    "head",
                    "tail"
                ],
                "skeleton": [
                    1,
                    2
                ]
            }
        ],
        "licenses": [
            {
                "name": "license1",
                "url": "example.com",
                "id": 1
            },
            {
                "name": "license2",
                "url": "example.com",
                "id": 2
            }
        ]
    }

    def test_coco_merge(self):
        merged = coco_merge(TestCocoMerge.example_coco_1, TestCocoMerge.example_coco_2)
        self.assertDictEqual(merged, TestCocoMerge.examples_merged)


import datetime
import json
from collections.abc import Iterable
import numpy as np
from PIL import Image
from . import utils

class COCOBase(object):
    '''
    base class to facilitate conversion of COCO stuff to a dictionary.

    TODO: refactor COCO classes to extend pycocotools' COCO

    TODO: refactor COCO classes to use AttrDict
    '''
    def _to_dict_fields(self, fields:list[str]) -> dict:
        return {field: self.__dict__[field] for field in fields if self.__dict__[field] is not None}
    def to_dict(self):
        '''convert COCO object to dictionary'''
        raise NotImplementedError('must implement a to_dict method!')


class COCOAnnotation(COCOBase):
    '''
        see https://cocodataset.org/#format-data

        see https://github.com/cocodataset/cocoapi/issues/184

            annotation {
                "id": int,
                "image_id": int,
                "category_id": int,
                "segmentation": RLE or [polygon],
                "area": float,
                "bbox": [x,y,width,height],
                "iscrowd": 0 or 1,
                "keypoints": [x1,y1,v1...],
                "num_keypoints": int
            }

        keypoints should be 3*len(keypoints_category), where keypoints_category is the keypoints in the corresponding category
    '''
    def __init__(self, image_id:int, eye_d:int, category_id:int, bbox:tuple[int]=None, area:float=None, segmentation=None, iscrowd:int=None, keypoints:list[int]=None):
        self.image_id = image_id
        self.id = eye_d
        self.category_id = category_id
        self.bbox = bbox
        area = area or self._compute_area()
        self.area = area
        self.segmentation = segmentation
        self.iscrowd = iscrowd or 0
        self.keypoints = keypoints
        self.num_keypoints = self._compute_num_keypoints()
            
    def _compute_num_keypoints(self):
        if self.keypoints is not None:
            # the number of keypoints is the number of "visible" keypoints.
            return sum([v > 0 for v in self.keypoints[::3]])
        return None

    def _compute_area(self):
        if self.bbox is not None:
            return self.bbox[-1] * self.bbox[-2]
        return None

    def to_dict(self):
        return self._to_dict_fields(
            ['image_id', 'id', 'category_id', 'bbox', 
             'area', 'segmentation', 'iscrowd', 
             'keypoints', 'num_keypoints'])
        

class COCOCategory(COCOBase):
    '''
        see https://cocodataset.org/#format-data
        
        see https://github.com/facebookresearch/Detectron/issues/640
        
            category {
                "id": int,
                "name": str,
                "supercategory": str,
                "keypoints": [str],
                "skeleton": [edge]
            }
        
        An edge is a tuple [a,b] where a,b are 1-indexed indices in the keypoints list.  So if keypoints is ["a", "b"], then [1,2] is an edge between "a" and "b"
    '''
    def __init__(self, name:str, eye_d:int, supercategory:str=None, keypoints:list[str]=None, skeleton:list[list[int]]=None):
        self.name = name
        self.id = eye_d
        self.supercategory = supercategory
        self.keypoints = keypoints
        self.skeleton = skeleton

    def to_dict(self):
        return self._to_dict_fields(
            ['name', 'id', 'supercategory', 'keypoints', 'skeleton']
        )


class COCOLicense(COCOBase):
    '''
        see https://cocodataset.org/#format-data
        
            license {
                "id": int,
                "name": str,
                "url": str,
            }
    '''
    def __init__(self, name:str, eye_d:int, url:str=None):
        self.name = name
        self.id = eye_d
        self.url = url

    def to_dict(self):
        return self._to_dict_fields(
            ['name', 'id', 'url']
        )


class COCOImage(COCOBase):
    '''
        see https://cocodataset.org/#format-data
        
            image {
                "id": int,
                "width": int,
                "height": int,
                "file_name": str,
                "license": int,
                "flickr_url": str,
                "coco_url": str,
                "date_captured": datetime,
            }
    '''
    def __init__(self, 
                 eye_d:int, file_name:str, 
                 width:int=None, height:int=None, 
                 license:int=None, coco_url:str=None, 
                 date_captured:datetime.datetime=None,
                 discover_image_properites=True):
        self.id = eye_d
        self.file_name = file_name
        self.width = width
        self.height = height
        if ((self.width is None) or (self.height is None)) and discover_image_properties:
            self.compute_width_and_height()
        self.license = license
        self.coco_url = coco_url
        self.date_captured = date_captured

    def compute_width_and_height(self):
        with Image.open(self.file_name) as im:
            self.width, self.height = im.size

    def to_dict(self) -> dict:
        the_dict = self._to_dict_fields(
            ['id', 'file_name', 'width', 'height', 'license', 'coco_url']
        )
        if self.date_captured:
            the_dict['date_captured'] = self.date_captured.isoformat()
        return the_dict

class COCOInfo(COCOBase):
    '''
        see https://cocodataset.org/#format-data
            
            info {
                "year": int,
                "version": str,
                "description": str,
                "contributor": str,
                "url": str,
                "date_created": datetime,
            }
    '''
    def __init__(self, year:int=None, version:str=None, description:str=None, contributor:str=None, url:str=None, date_created:datetime.datetime=None):
        self.year = year
        self.version = version
        self.description = description
        self.contributor = contributor
        self.url = url
        self.date_created = date_created

    def to_dict(self) -> dict:
        the_dict = self._to_dict_fields(
            ['year', 'version', 'description', 'contributor', 'url']
        )
        if self.date_created:
            the_dict['date_created'] = self.date_created.isoformat()
        return the_dict



class COCOCategories(object):
    '''
    helper class to hold the index on categories so that we can find categories by name or by index

    Parameters
    ----------
    categories: list[COCOCategory]
        existing list of COCOCategories if available

    Attributes
    ----------
    categories: list[COCOCategory]
        a list of unique categories
    category_map: dict[str, COCOCategory]
        maps category names to categories
    '''
    def __init__(self, categories: list[COCOCategory] | None = None):
        categories = categories or []
        # category ids MUST match their index+1 in the category list!
        self.categories = categories
        for i, category in enumerate(self.categories):
            assert category.id == i+1
        self.category_map = {category.name: category.id for category in self.categories}

    def add(self, label:str, keypoints:list[str]=None, skeleton:list[list[int]]=None) -> COCOCategory:
        '''
        Add a new category to the list.  Updates the map as well

        Parameters
        ----------
        label: str
            the string name of this category
        keypoints: list[str]
            the list of keypoint names for this category, if applicable
        skeleton: list[list[int]]
            the skeleton for this category, if applicable

        Returns
        -------
        category: COCOCategory
            returns the built COCOCategory
        '''
        if label not in self.category_map:
            category = COCOCategory(label, len(self.categories)+1, keypoints=keypoints, skeleton=skeleton)
            self.categories.append(category)
            self.category_map[self.categories[-1].name] = self.categories[-1].id
        return self.category_map[label]

    def __len__(self):
        return len(self.categories)

class COCOImages(object):
    '''
    helper class to hold the index on images so that we can find images by name or by index

    Parameters
    ----------
    images: list[COCOImage]
        existing list of COCOImages if available

    Attributes
    ----------
    images: list[COCOImage]
        a list of unique images
    image_map: dict[str, COCOImage]
        maps image names to images
    '''
    def __init__(self, images: list[COCOImage] | None = None):
        images = images or []
        # image ids MUST match their index+1 in the image list!
        self.images = images
        for i, image in enumerate(self.images):
            assert image.id == i+1
        self.image_map = {image.filename: image.id for image in self.images}

    def add(self, filename:str, width:int=None, height:int=None, 
            url:str=None, license:int=None, date_captured:datetime.datetime=None):
        '''
        Add a new image to the list.  Updates the map as well

        Parameters
        ----------
        filename: str
            the filename of this image
        width: int
            the width of this image, if known
        height: int
            the height of this image, if known
        url: str
            the url at which this image can be downloaded
        license: int
            the index in the list of COCOLicense with the applicable license
        date_captured: datetime.datetime
            the date and time when the image was captured

        Returns
        -------
        image: COCOImage
            returns the built COCOImage
        '''
        if filename not in self.image_map:
            image = COCOImage(len(self.images)+1, filename, 
                              width=width, height=height, coco_url=url, 
                              license=license, date_captured=date_captured)
            self.images.append(image)
            self.image_map[self.images[-1].file_name] = self.images[-1].id
        return self.image_map[filename]

    def __len__(self):
        return len(self.images)


class COCOData(COCOBase):
    '''
        see https://cocodataset.org/#format-data
        
            coco {
                "info": info,
                "images": [image],
                "annotations": [annotation],
                "licenses": [license],
                "categories": [category]
            }
    '''
    
    def __init__(self, info:COCOInfo, images:list[COCOImage], annotations:list[COCOAnnotation], licenses:list[COCOLicense], categories:list[COCOCategory]):
        self.info = info
        self.images = images
        self.annotations = annotations
        self.licenses = licenses
        self.categories = categories

    def to_dict(self) -> dict:
        return {
            'info': self.info.to_dict(),
            'images': [image.to_dict() for image in self.images],
            'annotations': [annotation.to_dict() for annotation in self.annotations],
            'licenses': [license.to_dict() for license in self.licenses],
            'categories': [category.to_dict() for category in self.categories]
        }

    def to_json(self, filename:str=None):
        '''
        dumps this COCO to json.  If a filename is provided, writes to disk, else, returns
        the string JSON

        Parameters
        ----------
        filename: str
            the optional location to which we should write the json
    
        Returns
        -------
        json: str | None
            if no filename is provided, returns the JSON COCO data as a string
        '''
        if filename is None:
            return json.dumps(self.to_dict(), cls=utils.NPEncoder)
        with open(filename, 'w') as f:
            json.dump(self.to_dict(), f, cls=utils.NPEncoder)
